`ifndef COMPUTE_UNIT_V
`define COMPUTE_UNIT_V

`include "VertSeek.v"
`include "VertSeekPipe.v"
`include "HorzSeek.v"
`include "Copier.v"
`include "ReadSwitch.v"
`include "DataStore.v"
`include "TapOrNot.v"

module ComputeUnit
#(
  parameter pipe_width_offset = 8'd1
)(
  input               clk,
  input               reset,

  input               compute_go,

  // Configuration
  input signed [15:0] bird_threshold,
  input [7:0]         bird_y,

  input signed [15:0] pipe_0_threshold,
  input [8:0]         pipe_scan_x,

  input [8:0]         bird_lower_limit,

  // Camera decoder m9k side
  output [16:0]       decoder_m9k_and_compute_addr,
  input [7:0]         decoder_m9k_and_compute_data,

  // VGA m9k side
  output [16:0]       compute_and_decoder_m9k_addr,
  output [7:0]        compute_and_decoder_m9k_data,
  output              compute_and_decoder_m9k_wren,

  // Output to seek m9k blocks
  output [9:0]        compute_and_bird_seek_m9k_data,
  output              compute_and_bird_seek_m9k_wren,

  output [9:0]        compute_and_horz_seek_m9k_data,
  output              compute_and_horz_seek_m9k_wren,

  output [9:0]        compute_and_vert_seek_m9k_data,
  output              compute_and_vert_seek_m9k_wren,

  output [8:0]        bird_threshold_display,
  output [8:0]        predicted_position,
  output [8:0]        bird_top_x_0,
  output [8:0]        bird_top_x_1,

  output [2:0]        tap_times,
  output reg          tap_times_val,

  output [8:0]        debug,
  output [8:0]        debug2
);

  reg                 reset_internal;
  reg                 reset_seeks;
  reg                 set_0_en;
  reg                 set_1_en;
  reg                 birdseek_go;

  assign debug2[0] = reset_seeks;
  assign debug2[1] = set_1_en;
  assign debug2[2] = set_0_en;
  assign debug2[5:3] = state;

  //======================================================================
  // Control
  //======================================================================

  reg [2:0]           state;
  reg [2:0]           state_next;

  localparam STATE_WAIT = 3'd0;
  localparam STATE_FRST = 3'd1;
  localparam STATE_DLAY = 3'd2;
  localparam STATE_SCND = 3'd3;
  localparam STATE_DONE = 3'd4;

  // State
  always @ ( posedge clk ) begin
    if ( reset ) begin
      state <= STATE_WAIT;
    end else begin
      state <= state_next;
    end
  end

  reg delay_countdown_reset;
  reg delay_countdown_dec;
  wire        countdown_done;

  vc_Counter
  #(
    .p_count_nbits       (24),
    .p_count_reset_value (24'd1000000),
    .p_count_max_value   (24'hfffff)
  )
  delay_countdown
  (
   .clk           (clk),
   .reset         (reset || delay_countdown_reset),
   .increment     (1'b0),
   .decrement     (delay_countdown_dec),
   .count         (),
   .count_is_zero (countdown_done),
   .count_is_max  ()
  );

  // State Transitions
  always @ ( * ) begin
    case ( state )
      STATE_WAIT: begin
        state_next <= compute_go ? STATE_FRST : STATE_WAIT;
      end
      STATE_FRST: begin
        if ( horzseek_done && vertseek_done && birdseek_done )
          state_next <= STATE_DLAY;
        else
          state_next <= STATE_FRST;
      end
      STATE_DLAY: begin
        if ( countdown_done )
          state_next <= STATE_SCND;
        else
          state_next <= STATE_DLAY;
      end
      STATE_SCND: begin
        if ( horzseek_done && vertseek_done && birdseek_done )
          state_next <= STATE_DONE;
        else
          state_next <= STATE_SCND;
      end
      STATE_DONE: begin
        state_next <= STATE_WAIT;
      end
    endcase
  end

  // State Outputs
  always @ ( * ) begin
    case ( state )
      STATE_WAIT: begin
        reset_internal <= 1'b0;
        reset_seeks <= 1'b0;
        set_0_en <= 1'b0;
        set_1_en <= 1'b0;
        tap_times_val <= 1'b0;
        birdseek_go <= 1'b0;
        delay_countdown_reset <= 1'b1;
        delay_countdown_dec <= 1'b0;
      end
      STATE_FRST: begin
        reset_internal <= 1'b0;
        reset_seeks <= ( horzseek_done && vertseek_done && birdseek_done );
        set_0_en <= 1'b1;
        set_1_en <= 1'b0;
        tap_times_val <= 1'b0;
        // Actually not exactly a go signal
        birdseek_go <= 1'b1;
        delay_countdown_reset <= 1'b1;
        delay_countdown_dec <= 1'b0;
      end
      STATE_DLAY: begin
        reset_internal <= 1'b0;
        reset_seeks <= 1'b0;
        set_0_en <= 1'b0;
        set_1_en <= 1'b0;
        tap_times_val <= 1'b0;
        // Actually not exactly a go signal
        birdseek_go <= 1'b0;
        delay_countdown_reset <= 1'b0;
        delay_countdown_dec <= 1'b1;
      end
      STATE_SCND: begin
        reset_internal <= 1'b0;
        reset_seeks <= ( horzseek_done && vertseek_done && birdseek_done );
        set_0_en <= 1'b0;
        set_1_en <= 1'b1;
        tap_times_val <= 1'b0;
        // Actually not exactly a go signal
        birdseek_go <= 1'b1;
        delay_countdown_reset <= 1'b1;
        delay_countdown_dec <= 1'b0;
      end
      STATE_DONE: begin
        reset_internal <= 1'b1;
        reset_seeks <= 1'b1;
        set_0_en <= 1'b0;
        set_1_en <= 1'b0;
        tap_times_val <= 1'b1;
        birdseek_go <= 1'b0;
        delay_countdown_reset <= 1'b1;
        delay_countdown_dec <= 1'b0;
      end
    endcase
  end

  //======================================================================
  // Datapath
  //======================================================================

  //----------------------------------------------------------------------
  // Read switch
  //----------------------------------------------------------------------

  wire         pause_read;

  wire [16:0]  read_mux_copier_addr;
  wire [7:0]   read_mux_copier_data;

  wire         bird_demand_access_read;
  wire [16:0]  read_mux_bird_addr;
  wire [7:0]   read_mux_bird_data;

  wire         horz_demand_access_read;
  wire [16:0]  read_mux_horz_addr;
  wire [7:0]   read_mux_horz_data;

  wire         vert_demand_access_read;
  wire [16:0]  read_mux_vert_addr;
  wire [7:0]   read_mux_vert_data;

  ReadSwitch read_switch
  (
   .data     (decoder_m9k_and_compute_data),
   .addr     (decoder_m9k_and_compute_addr),

   .pause_c  (pause_read),
   .data_c   (read_mux_copier_data),
   .addr_c   (read_mux_copier_addr),

   .demand_0 (bird_demand_access_read),
   .data_0   (read_mux_bird_data),
   .addr_0   (read_mux_bird_addr),

   .demand_1 (horz_demand_access_read),
   .data_1   (read_mux_horz_data),
   .addr_1   (read_mux_horz_addr),

   .demand_2 (vert_demand_access_read),
   .data_2   (read_mux_vert_data),
   .addr_2   (read_mux_vert_addr)
  );

  //--------------------------------------------------------------------
  // Copier
  //--------------------------------------------------------------------

  Copier copier
  (
   .clk                   (clk),
   .reset                 (reset),

   .pause                 (pause_read),

   .decoder_side_m9k_addr (read_mux_copier_addr),
   .decoder_side_m9k_data (read_mux_copier_data),

   .vga_side_m9k_addr     (compute_and_decoder_m9k_addr),
   .vga_side_m9k_data     (compute_and_decoder_m9k_data),
   .vga_side_m9k_wren     (compute_and_decoder_m9k_wren)
  );

  //--------------------------------------------------------------------
  // Birdseek
  //--------------------------------------------------------------------

  wire [8:0]    bird_top_x_in;
  wire          bird_top_x_wren;

  wire          birdseek_done;

  VertSeek bird_seek
  (
   .clk                   (clk),
   .reset                 (reset || reset_seeks),

   .pause                 (1'b0),
   .go                    (birdseek_go),

   .threshold             (bird_threshold),
   .y                     (bird_y),

   .decoder_side_m9k_addr (read_mux_bird_addr),
   .decoder_side_m9k_data (read_mux_bird_data),

   .identified_x          (bird_top_x_in),
   .identified_x_wren     (bird_top_x_wren),

   .done                  (birdseek_done),

   .demand_access_read    (bird_demand_access_read)
  );

  assign compute_and_bird_seek_m9k_data = { bird_top_x_wren, bird_top_x_in };
  assign compute_and_bird_seek_m9k_wren = bird_top_x_wren;

  //--------------------------------------------------------------------
  // HorzSeek
  //--------------------------------------------------------------------

  wire          pipe_0_y_wren;
  wire [7:0]    pipe_0_y_in;

  wire          horzseek_done;

  HorzSeek pipe_0_seek
  (
   .clk                   (clk),
   .reset                 (reset || reset_seeks),

   .pause                 (1'b0),
   .go                    (birdseek_done),

   .scan_start_y          (bird_y),
   .threshold             (pipe_0_threshold),

   .x                     (pipe_scan_x),

   .decoder_side_m9k_addr (read_mux_horz_addr),
   .decoder_side_m9k_data (read_mux_horz_data),

   .identified_y          (pipe_0_y_in),
   .identified_y_wren     (pipe_0_y_wren),

   .done                  (horzseek_done),

   .demand_access_read    (horz_demand_access_read)
  );

  assign compute_and_horz_seek_m9k_data = { 1'b0, pipe_0_y_wren, pipe_0_y_in };
  assign compute_and_horz_seek_m9k_wren = pipe_0_y_wren;

  //--------------------------------------------------------------------
  // VertSeek
  //--------------------------------------------------------------------

  wire          pipe_0_gap_top_x_wren;
  wire [8:0]    pipe_0_gap_top_x_in;

  wire [7:0]    pipe_0_y = set_1_en ? pipe_0_y_1 : pipe_0_y_0;

  wire          vertseek_done;

  VertSeekPipe pipe_0_vert_seek
  (
   .clk                   (clk),
   .reset                 (reset || reset_seeks),

   .pause                 (1'b0),
   .go                    (horzseek_done),

   .threshold             (pipe_0_threshold),
   .y                     (pipe_0_y - pipe_width_offset),

   .decoder_side_m9k_addr (read_mux_vert_addr),
   .decoder_side_m9k_data (read_mux_vert_data),

   .identified_x          (pipe_0_gap_top_x_in),
   .identified_x_wren     (pipe_0_gap_top_x_wren),

   .done                  (vertseek_done),

   .demand_access_read    (vert_demand_access_read)
  );

  assign compute_and_vert_seek_m9k_data = { pipe_0_gap_top_x_wren, pipe_0_gap_top_x_in };
  assign compute_and_vert_seek_m9k_wren = pipe_0_gap_top_x_wren;

  //----------------------------------------------------------------------
  // Data Stores
  //----------------------------------------------------------------------

  // The 0 set

  wire          bird_top_x_0_valid;

  DataStore #(9) bird_top_x_datastore_0
  (
    .clk      (clk),
    .reset    (reset),
    .clear    (reset_internal),
    .en       (set_0_en),
    .valid_in (bird_top_x_wren),
    .data_in  (bird_top_x_in),
    .valid    (bird_top_x_0_valid),
    .data     (bird_top_x_0)
  );

  wire          pipe_0_y_0_valid;
  wire [7:0]    pipe_0_y_0;

  DataStore #(8) pipe_0_y_datastore_0
  (
    .clk      (clk),
    .reset    (reset),
    .clear    (reset_internal),
    .en       (set_0_en),
    .valid_in (pipe_0_y_wren),
    .data_in  (pipe_0_y_in),
    .valid    (pipe_0_y_0_valid),
    .data     (pipe_0_y_0)
  );

  wire          pipe_0_gap_top_x_0_valid;
  wire [8:0]    pipe_0_gap_top_x_0;

  DataStore #(9) pipe_0_gap_top_x_datastore_0
  (
    .clk      (clk),
    .reset    (reset),
    .clear    (reset_internal),
    .en       (set_0_en),
    .valid_in (pipe_0_gap_top_x_wren),
    .data_in  (pipe_0_gap_top_x_in),
    .valid    (pipe_0_gap_top_x_0_valid),
    .data     (pipe_0_gap_top_x_0)
  );

  // The 1 set

  wire          bird_top_x_1_valid;

  DataStore #(9) bird_top_x_datastore_1
  (
    .clk      (clk),
    .reset    (reset),
    .clear    (reset_internal),
    .en       (set_1_en),
    .valid_in (bird_top_x_wren),
    .data_in  (bird_top_x_in),
    .valid    (bird_top_x_1_valid),
    .data     (bird_top_x_1)
  );

  wire          pipe_0_y_1_valid;
  wire [7:0]    pipe_0_y_1;

  DataStore #(8) pipe_0_y_datastore_1
  (
    .clk      (clk),
    .reset    (reset),
    .clear    (reset_internal),
    .en       (set_1_en),
    .valid_in (pipe_0_y_wren),
    .data_in  (pipe_0_y_in),
    .valid    (pipe_0_y_1_valid),
    .data     (pipe_0_y_1)
  );

  wire          pipe_0_gap_top_x_1_valid;
  wire [8:0]    pipe_0_gap_top_x_1;

  DataStore #(9) pipe_0_gap_top_x_datastore_1
  (
    .clk      (clk),
    .reset    (reset),
    .clear    (reset_internal),
    .en       (set_1_en),
    .valid_in (pipe_0_gap_top_x_wren),
    .data_in  (pipe_0_gap_top_x_in),
    .valid    (pipe_0_gap_top_x_1_valid),
    .data     (pipe_0_gap_top_x_1)
  );

  //--------------------------------------------------------------------
  // TapOrNot?
  //--------------------------------------------------------------------

  TapOrNot tap_algorithm
  (
   .bird_lower_limit         (bird_lower_limit),

   // Inputs from scanning
   .bird_top_x_0_valid       (bird_top_x_0_valid),
   .bird_top_x_0             (bird_top_x_0),

   .pipe_0_y_0_valid         (pipe_0_y_0_valid),
   .pipe_0_y_0               (pipe_0_y_0),

   .pipe_0_gap_top_x_0_valid (pipe_0_gap_top_x_0_valid),
   .pipe_0_gap_top_x_0       (pipe_0_gap_top_x_0),

   // Inputs from scanning
   .bird_top_x_1_valid       (bird_top_x_1_valid),
   .bird_top_x_1             (bird_top_x_1),

   .pipe_0_y_1_valid         (pipe_0_y_1_valid),
   .pipe_0_y_1               (pipe_0_y_1),

   .pipe_0_gap_top_x_1_valid (pipe_0_gap_top_x_1_valid),
   .pipe_0_gap_top_x_1       (pipe_0_gap_top_x_1),

   .bird_threshold_display   (bird_threshold_display),

   .tap_times                (tap_times),
   .predicted_position       (predicted_position),

   .debug                    ()
  );

endmodule

`endif
