`include "vc-Counter.v"

module Tapper
#(
  parameter count_max = 24'd2000000
)(
  input        clk,
  input        reset,

  // Configuration
  input [23:0] num_tap_cycles,
  input [23:0] num_between_tap_cycles,

  input [2:0]  tap_times,
  // Can also be used as the go signal
  input        tap_times_val,

  output reg   tap_out,
  output       done
);

  // Register inputs

  reg [2:0]   tap_times_internal;
  reg         update;
  always @ ( posedge clk ) begin
    if ( reset ) begin
      tap_times_internal <= 3'd0;
    end else begin
      if ( update )
        tap_times_internal <= tap_times_internal - 1'b1;
      else if ( tap_times_val && state == STATE_WAIT )
        tap_times_internal <= tap_times;
    end
  end

  // Tapping stuff

  wire [23:0] count;

  reg         increment;
  reg         reset_counter;

  vc_Counter
  #(
    .p_count_nbits       (24),
    .p_count_reset_value (24'd0),
    .p_count_max_value   (24'hffffff)
  )
  counter
  (
   .clk           (clk),
   .reset         (reset || reset_counter),
   .increment     (increment),
   .decrement     (1'b0),
   .count         (count),
   .count_is_zero (),
   .count_is_max  ()
  );

  reg [1:0]        state;
  reg [1:0]        state_next;

  localparam STATE_WAIT          = 2'd0;
  localparam STATE_TAP           = 2'd1;
  localparam STATE_BETWEEN_TAPS  = 2'd2;
  localparam STATE_DONE          = 2'd3;

  always @ ( posedge clk ) begin
    if ( reset ) begin
      state <= STATE_WAIT;
    end else begin
      state <= state_next;
    end
  end

  // State transitions
  always @ ( * ) begin
    case (state)
      STATE_WAIT: begin
        state_next <= tap_times_val && tap_times > 'b0 ? STATE_TAP : STATE_WAIT;
      end
      STATE_TAP: begin
        if ( count == num_tap_cycles ) begin
          if ( tap_times_internal == 3'd1 ) begin
            state_next <= STATE_DONE;
          end else begin
            state_next <= STATE_BETWEEN_TAPS;
          end
        end else begin
          state_next <= STATE_TAP;
        end
      end
      STATE_BETWEEN_TAPS: begin
        state_next <= ( count == num_between_tap_cycles ) ? STATE_TAP : STATE_BETWEEN_TAPS;
      end
      STATE_DONE: begin
        state_next <= STATE_DONE;
      end
    endcase
  end

  // State outputs
  always @ ( * ) begin
    case (state)
      STATE_WAIT: begin
        increment <= 1'b0;
        reset_counter <= 1'b1;
        update <= 1'b0;
        tap_out <= 1'b0;
      end
      STATE_TAP: begin
        increment <= 1'b1;
        reset_counter <= ( count == num_tap_cycles );
        update <= ( count == num_tap_cycles );
        tap_out <= 1'b1;
      end
      STATE_BETWEEN_TAPS: begin
        increment <= 1'b1;
        reset_counter <= ( count == num_between_tap_cycles );
        update <= 1'b0;
        tap_out <= 1'b0;
      end
      STATE_DONE: begin
        increment <= 1'b0;
        reset_counter <= 1'b1;
        update <= 1'b0;
        tap_out <= 1'b0;
      end
    endcase
  end

  assign done = state == STATE_DONE;

endmodule
