`ifndef VERT_SEEK_PIPE_V
`define VERT_SEEK_PIPE_V

`include "SignedMult.v"
`include "vc-regs.v"

module VertSeekPipe
(
 input               clk,
 input               reset,

 input               pause,
 input               go,

 input signed [15:0] threshold,

 // Remember that coordinates are rotated since the phone is on its side
 input [7:0]         y,

 output reg [16:0]   decoder_side_m9k_addr,
 input [7:0]         decoder_side_m9k_data,

 // identified_x holds the first x pos where the gradient is greater
 // than a threshold
 output [8:0]        identified_x,
 output              identified_x_wren,

 output reg          done,

 output reg          demand_access_read
);

  //----------------------------------------------------------------------
  // Past two entries
  //----------------------------------------------------------------------
  localparam LOAD_ZERO    = 2'd0;
  localparam LOAD_ONE     = 2'd1;
  localparam LOAD_BOTH    = 2'd2;
  localparam LOAD_DISABLE = 2'd3;

  reg [15:0]       past_two;
  reg [1:0]        past_two_load;
  reg [15:0]       past_two_next;
  wire [15:0]      past_two_shift = { past_two[7:0], decoder_side_m9k_data };

  always @ ( posedge clk ) begin
    if ( reset ) begin
      past_two <= 16'b0;
    end else begin
      past_two <= pause ? past_two : past_two_next;
    end
  end

  always @ ( * ) begin
    case ( past_two_load )
      LOAD_ZERO:    past_two_next = { past_two[15:8], decoder_side_m9k_data };
      LOAD_ONE:     past_two_next = { decoder_side_m9k_data, past_two[7:0] };
      LOAD_BOTH:    past_two_next = past_two_shift;
      LOAD_DISABLE: past_two_next = past_two;
    endcase
  end

  //----------------------------------------------------------------------
  // Computation stuff (combinational)
  //----------------------------------------------------------------------
  // The concept here is edge detection using correlation with
  // thresholding
  // In two's complement: +1, 0, -1
  wire [7:0]       kernel [2:0];
  assign kernel[0] = 8'h01;
  assign kernel[1] = 8'h00;
  assign kernel[2] = 8'hff;

  wire [7:0]       half_max_intensity = 8'd127;

  wire [15:0]      mult_out [2:0];
  SignedMult #(8) mult_0
  (
   .a   (kernel[0]),
   .b   (past_two[15:8] - half_max_intensity),
   .out (mult_out[0])
  );

  SignedMult #(8) mult_1
  (
   .a   (kernel[1]),
   .b   (past_two[7:0] - half_max_intensity),
   .out (mult_out[1])
  );

  SignedMult #(8) mult_2
  (
   .a   (kernel[2]),
   .b   (decoder_side_m9k_data - half_max_intensity),
   .out (mult_out[2])
  );

  wire signed [15:0] correlation = mult_out[0] + mult_out[1] + mult_out[2];

  //----------------------------------------------------------------------
  // Internal x value for m9k request
  //----------------------------------------------------------------------
  wire [8:0]       x;
  reg              x_en;
  wire [8:0]       x_next = x - 1'b1;

  vc_EnResetReg #(9,9'd380) x_internal
  (
   .clk   (clk),
   .reset (reset),
   .en    (x_en && !pause),
   .q     (x),
   .d     (x_next)
  );

  //----------------------------------------------------------------------
  // State definitions
  //----------------------------------------------------------------------
  localparam STATE_WAIT         = 3'd0;
  localparam STATE_PRE_LOAD     = 3'd1;
  localparam STATE_LOAD_INITIAL = 3'd2;
  localparam STATE_PROCESS      = 3'd3;
  localparam STATE_DONE         = 3'd4;

  // State
  reg [2:0]    state;
  reg [2:0]    state_next;

  always @ ( posedge clk ) begin
    if ( reset ) begin
      state <= STATE_WAIT;
    end else begin
      state <= pause ? state : state_next;
    end
  end

  wire identified_x_valid;

  // State Transitions
  always @ ( * ) begin
    case ( state )
      STATE_WAIT: begin
        state_next <= go ? STATE_PRE_LOAD : STATE_WAIT;
      end
      STATE_PRE_LOAD: begin
        state_next <= STATE_LOAD_INITIAL;
      end
      STATE_LOAD_INITIAL: begin
        if ( x == 9'd379 )
          // Allow first two entries in memory to load
          state_next <= STATE_PROCESS;
        else
          state_next <= STATE_LOAD_INITIAL;
      end
      STATE_PROCESS: begin
        if ( x == 9'd0 || identified_x_valid )
          state_next <= STATE_DONE;
        else
          state_next <= STATE_PROCESS;
      end
      STATE_DONE: begin
        state_next <= STATE_DONE;
      end
    endcase
  end

  // Outputs
  always @ ( * ) begin
    case ( state )
      STATE_WAIT: begin
        decoder_side_m9k_addr <= 17'b0;
        past_two_load <= LOAD_DISABLE;
        x_en <= 1'b0;
        done <= 1'b0;
        demand_access_read <= 1'b0;
      end
      STATE_PRE_LOAD: begin
        decoder_side_m9k_addr <= { y, x };
        past_two_load <= LOAD_DISABLE;
        x_en <= 1'b1;
        done <= 1'b0;
        demand_access_read <= 1'b1;
      end
      STATE_LOAD_INITIAL: begin
        decoder_side_m9k_addr <= { y, x };
        past_two_load <= ( x == 9'd380 ) ? LOAD_ZERO : LOAD_ONE;
        x_en <= 1'b1;
        done <= 1'b0;
        demand_access_read <= 1'b1;
      end
      STATE_PROCESS: begin
        decoder_side_m9k_addr <= { y, x };
        past_two_load <= LOAD_BOTH;
        x_en <= 1'b1;
        done <= 1'b0;
        demand_access_read <= 1'b1;
      end
      STATE_DONE: begin
        decoder_side_m9k_addr <= 17'b0;
        past_two_load <= LOAD_DISABLE;
        x_en <= 1'b0;
        done <= 1'b1;
        demand_access_read <= 1'b0;
      end
    endcase
  end

  assign identified_x = x;
  assign identified_x_valid = correlation > threshold;
  assign identified_x_wren = identified_x_valid && state == STATE_PROCESS;

endmodule

`endif
